@import url(https://fonts.googleapis.com/css2?family=Roboto&family=Roboto+Mono&display=swap);@import url(https://fonts.googleapis.com/icon?family=Material+Icons);@font-face{font-display:"swap";font-family:UCity;font-weight:300;src:local("UCity Light"),url(/static/media/UCity-Light.30446c15a21bf8a994e8.woff2) format("woff2"),url(/static/media/UCity-Light.47cf3e33cb6bb1496c95.woff) format("woff")}@font-face{font-display:"swap";font-family:UCity;font-weight:400;src:local("UCity Regular"),url(/static/media/UCity-Regular.8afa44ac39706d62b62b.woff2) format("woff2"),url(/static/media/UCity-Regular.7a3b475525ed92dc7816.woff) format("woff")}@font-face{font-display:"swap";font-family:UCity;font-weight:600;src:local("UCity Semibold"),url(/static/media/UCity-Semibold.80655a0e8b4bb5f30545.woff2) format("woff2"),url(/static/media/UCity-Semibold.bb7aab96e378ebd6a651.woff) format("woff")}html{overflow-x:hidden}body,html{height:100%;width:100%}body{-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale;font-family:-apple-system,BlinkMacSystemFont,Segoe UI,Roboto,Oxygen,Ubuntu,Cantarell,Fira Sans,Droid Sans,Helvetica Neue,sans-serif;margin:0;overflow-y:overlay}#root{height:100%;width:100%}code{font-family:source-code-pro,Menlo,Monaco,Consolas,Courier New,monospace}
/*# sourceMappingURL=main.bb0f092c.css.map*/