.App {
  text-align: center;
	height: 100vh;
	background: rgb(34,193,195);
	background: linear-gradient(0deg, rgba(34,193,195,1) 0%, rgba(253,187,45,1) 100%);
}

.App .wrapper{
	transform: translateY(100%);
}
